----------------------------------------------------------------
--
--  Copyright 1989, Andy Tsay, and Alec Stanculescu, and 
--            1990, Fintronic USA, Inc.
--
--  Note: The code in this file has been developed 
--        by Andy Tsay and Alec Stanculescu at 
--        Vantage Analysis Systems, Inc., and 
--        Fintronic USA, Inc., and modified in minor aspects
--        in 1990 by Alec Stanculescu at Fintronic USA, Inc.
--
--        Vantage Analysis Systems, Inc. has graciously
--        placed the Std_Logic Package in the public domain.
--
--        The only restrictions in using this code are:
--        1. Observe the export regulations of the United States
--           of America. When in doubt check with the Department
--           of Commerce, with Vantage Analysis Systems, Inc, or with 
--           Alec Stanculescu (Fax 415 578 0260).
--        2. Do not separate the Copyright notice nor this note 
--           from the code.
--        3. If you modify the code, give proper credit to this 
--           version by acknowledging: Vantage Analysis Systems,
--           Fintronic USA, Andy Tsay, and Alec Stanculescu.
--
----------------------------------------------------------------
--                            switch package
----------------------------------------------------------------
--
USE work.STD_LOGIC.ALL;
-------------------------------
-- bxfr: Bidirectional XFR gate
-------------------------------
--
entity bxfr is
    generic (gdelay : time := 3 ps; MaxStrength : t_strength := 'R');
         --  gdelay      : delay on the gate G
         --  MaxStrength : max output strength on Src and Drn. Used to model resistance.
    port( g: in t_wlogic; src,drn: inout t_wlogic);
         -- Src and Drn are symmetric.
end bxfr;
--
architecture bxfr_behavior of bxfr is
        signal wakeup : boolean := false;
        signal gd : t_wlogic;
begin
    gd <= g after gdelay;
    process(gd,src'transaction,drn'transaction)
        variable psrc, pdrn : t_logic;   -- to store old  contributions
        variable vsrc, vdrn : t_logic;   -- to store next contributions
        variable init : boolean := true; -- to initialize src and drn at powerup
    begin                         ----------------------
        if not wakeup then        ----- disconnect -----
            if init then          ----------------------
                src <= D;
                drn <= D;
                init := false;
            elsif (gd'event or (f_state(gd) /= '0')) then
                    wakeup <= true;
                    psrc := f_convz(src);
                    pdrn := f_convz(drn);
                    src <= psrc;
                    drn <= pdrn;

            else -- no event on gate, and gate is off
                if src'active then  -- kick the src back
                    vsrc := f_convz(src);
                    if psrc /= vsrc then  -- update when change
                       psrc := vsrc;
                       if f_strength(src) /= 'F' then
                          src <= vsrc;
                       end if;
                    end if;
                end if;
                if drn'active then  -- kick the drn back
                    vdrn := f_convz(drn);
                    if pdrn /= vdrn then  -- update when change
                       pdrn := vdrn;
                       if f_strength(drn) /= 'F' then
                          drn <= vdrn;
                       end if;
                    end if;
                end if;
            end if;
        else                      ----------------------
            case f_state(gd) is   ----  reconnect  -----
                when '0' =>       ----------------------
                                  -- f_convz computes tristate value
                    vsrc := f_convz(src);
                    vdrn := f_convz(drn);
                                  
                when '1' =>       -- compute brf of src and drn
                                  -- f_ceil computes the resistance
                                  -- f_busres computes the resolved value
                    vsrc := f_ceil(f_busres(src)(drn),MaxStrength);
                    vdrn := f_ceil(f_busres(src)(drn),MaxStrength);

                when 'X' =>       -- compute range between tristate and brf of src and drn
                                  -- f_uxfr is the function for unidirectional transistor
                                  -- use f_uxfr to compute the range between tristate
                                  --     and resolved value
                    vsrc := f_ceil(f_uxfr(FX,f_busres(src)(drn),f_convz(src)),MaxStrength);
                    vdrn := f_ceil(f_uxfr(FX,f_busres(src)(drn),f_convz(drn)),MaxStrength);
            end case;
            if  (psrc = vsrc) and (pdrn = vdrn) then -- wakeup in 1 fs, if nothing changes
                    wakeup <= false after 1 fs;
            else
               if psrc /= vsrc then  -- update when change
                   psrc := vsrc;
                   if f_strength(src) /= 'F' then
                      src <= vsrc;
                   end if;
               end if;
               if pdrn /= vdrn then  -- update when change
                   pdrn := vdrn;
                   if f_strength(drn) /= 'F' then
                      drn <= vdrn;
                   end if;
               end if;
            end if;
        end if;
    end process;
end bxfr_behavior;

---------------------------
-- nfet
---------------------------
USE work.STD_LOGIC.ALL;
USE work.ALL;
entity nfet is
    generic (gdelay : time := 3 ps; MaxStrength : t_strength := 'R');
    port( g: in t_wlogic; src,drn: inout t_wlogic);
end nfet;
architecture nfet_behavior of nfet is
    component bxfr
        generic (gdelay : time := 3 ps; MaxStrength : t_strength := 'R');
        port( g: in t_wlogic; src,drn: inout t_wlogic); end component;
begin
    i1: bxfr generic map (gdelay, MaxStrength) port map( g, src, drn );
end nfet_behavior;

USE work.ALL;

---------------------------
-- pfet
---------------------------
USE work.STD_LOGIC.ALL;
USE work.ALL;
entity pfet is
    generic (gdelay : time := 3 ps; MaxStrength : t_strength := 'R');
    port( g: in t_wlogic; src,drn: inout t_wlogic);
end pfet;
architecture pfet_behavior of pfet is
    component bxfr 
        generic (gdelay : time := 3 ps; MaxStrength : t_strength := 'R');
        port( g: in t_wlogic; src,drn: inout t_wlogic); end component;
    signal tg : t_wlogic;
begin
    tg <= f_logic(f_not(f_state(g)))('R');
    i1: bxfr generic map (gdelay, MaxStrength) port map( tg, src, drn );
end pfet_behavior;
