
--------------------------------------------------------------------------
--                                     copyright 1988
--                                     Vantage Analysis Systems, Inc.
--                                       all rights reserved
--
--   File name :  std_logic.vhdl
--   Title     :  STD_LOGIC package VHDL source
--   Subsystem :  analyzer
--   Module    :  VHDL source
--
--   Purpose   :  source for STD_LOGIC package
--
--   Author(s) :  dlp,kes,drc,ast
--
--   Revision  :  1/14/89
--                1/20/89
--
--   Remarks   :  Defines the TYPEs for the standard logic value
--                system and also some FUNCTIONs and arrays to access parts
--                of a standard logic value. Also provided are standard logic
--                FUNCTIONs on these logic values.
--
--------------------------------------------------------------------------

--##################################################################################
--##################################################################################
--##################                                              ##################
--##################               PACKAGE  INTERFACE             ##################
--##################                                              ##################
--##################################################################################
--##################################################################################

PACKAGE std_logic IS

    --$ !VANTAGE_METACOMMENTS_ON
    --$ !VANTAGE_DNA_ON

    -- ***************************************************************************
    --
    --                        PUBLIC TYPE DECLARATIONS.
    --
    -- ***************************************************************************
    
    ------------------------------------------------------------------------------
    --
    -- t_logic:  std-value TYPE used IN simulation.
    --
    ------------------------------------------------------------------------------
    TYPE t_logic IS (
        U,  D,
        Z0, Z1, ZDX, DZX, ZX,
        W0, W1, WZ0, WZ1, WDX, DWX, WZX, ZWX, WX,
        R0, R1, RW0, RW1, RZ0, RZ1, RDX, DRX, RZX, ZRX, RWX, WRX, RX,
        F0, F1, FR0, FR1, FW0, FW1, FZ0, FZ1, FDX, DFX, FZX, ZFX, FWX, WFX, FRX, RFX, FX
        );


    ------------------------------------------------------------------------------
    --
    -- t_logic_vector:  An array of std-state values.
    --
    ------------------------------------------------------------------------------

    TYPE t_logic_vector IS ARRAY(NATURAL RANGE <>) OF t_logic;


    ------------------------------------------------------------------------------
    --
    -- f_logic_bus:  Resolution FUNCTION to be used on all SIGNALs of TYPE t_wlogic.
    --
    ------------------------------------------------------------------------------

    FUNCTION f_logic_bus( s : t_logic_vector ) RETURN t_logic;


    ------------------------------------------------------------------------------
    --
    -- t_wlogic:  std-state TYPE declaration to be used with all SIGNAL declarations.
    --
    ------------------------------------------------------------------------------

    SUBTYPE t_wlogic IS f_logic_bus t_logic;


    ------------------------------------------------------------------------------
    --
    -- t_wlogic_vector:  An array of std-state values.  To be used IN declaring
    --                   SIGNAL arrays of std-state TYPEs.
    --
    ------------------------------------------------------------------------------

    TYPE t_wlogic_vector IS ARRAY(NATURAL RANGE <>) OF t_wlogic;


    ------------------------------------------------------------------------------
    --
    -- t_state:  The three basic state values.
    --
    ------------------------------------------------------------------------------

    TYPE t_state IS ( '0', '1', 'X' );


    ------------------------------------------------------------------------------
    --
    -- t_strength:  The different strengths, IN ascENDing order.
    --
    ------------------------------------------------------------------------------

    TYPE t_strength IS ( 'U', 'Z', 'W', 'R', 'F' );


    ------------------------------------------------------------------------------
    --
    -- t_technology:  The different technologies supported by this package.
    --
    ------------------------------------------------------------------------------
    
    TYPE t_technology IS ( ecl, cmos, nmos, ttl, ttloc );

    ------------------------------------------------------------------------------
    --
    -- Internal Types:  Types for look-up tables:
    --     f_state_T      : Takes t_logic value, RETURNs the t_state.
    --     f_strength_T   : Takes t_logic value, RETURNs the t_strength.
    --     f_logic_T      : Takes t_logic value, converts to different t_logic.
    --     
    --     F_1_X_1        : Takes t_state, converts to a different t_state.  
    --                      (ie, f_convz, f_not)
    --     F_2_X_1        : Takes 2 t_state values, converts to new t_state.  
    --                      (ie, f_and, f_or, f_nand, f_nor, f_xor)
    --     
    --     f_str          : Used IN f_log_con_t.
    --     f_log_con_t    : Takes t_state and t_strength, RETURNs t_logic.
    --     
    --     f_tech_con     : Used IN f_tech_T.
    --     f_tech_T       : Takes t_state and t_technology, RETURNs t_logic.
    --     f_specific_T   : Takes t_state, RETURNs t_logic based on technology.
    --
    --     f_boolean_T    : Takes t_logic, RETURNs boolean.
    --     
    ------------------------------------------------------------------------------
    
    TYPE f_state_T IS ARRAY (t_logic'low to t_logic'high) OF t_state;                                
    TYPE f_strength_T IS ARRAY (t_logic'low to t_logic'high) OF t_strength;
    TYPE f_logic_T IS ARRAY (t_logic'low to t_logic'high) OF t_logic;

    TYPE F_1_X_1 IS ARRAY (t_state'low to t_state'high) OF t_state;
    TYPE F_2_X_1 IS ARRAY (t_state'low to t_state'high) OF F_1_X_1;

    TYPE f_str IS ARRAY (t_strength'low to t_strength'high) OF t_logic;
    TYPE f_log_con_T IS ARRAY (t_state'low to t_state'high) OF f_str;
    TYPE f_logs_con_t IS ARRAY (t_strength'LOW TO t_strength'HIGH) OF f_str;

    TYPE f_tech_con IS ARRAY (t_technology'low to t_technology'high) OF t_logic;
    TYPE f_tech_T IS ARRAY (t_state'low to t_state'high) OF f_tech_con;
    TYPE f_specific_t IS array( t_state'low to t_state'high ) of t_logic;

    TYPE f_boolean_t IS array( t_logic'low to t_logic'high ) of boolean; 

    -- ***************************************************************************
    --
    --             PUBLIC OVERLOADED COMPARISON OPERATORS DECLARATIONS.
    --
    -- ***************************************************************************
    
    FUNCTION "="  ( l : t_logic;  r : t_state )  RETURN boolean;
    FUNCTION "="  ( l : t_state;  r : t_logic )  RETURN boolean;
    FUNCTION "/=" ( l : t_logic;  r : t_state )  RETURN boolean;
    FUNCTION "/=" ( l : t_state;  r : t_logic )  RETURN boolean;
    FUNCTION ">"  ( l : t_logic;  r : t_state )  RETURN boolean;
    FUNCTION ">"  ( l : t_state;  r : t_logic )  RETURN boolean;
    FUNCTION ">=" ( l : t_logic;  r : t_state )  RETURN boolean;
    FUNCTION ">=" ( l : t_state;  r : t_logic )  RETURN boolean;
    FUNCTION "<"  ( l : t_logic;  r : t_state )  RETURN boolean;
    FUNCTION "<"  ( l : t_state;  r : t_logic )  RETURN boolean;
    FUNCTION "<=" ( l : t_logic;  r : t_state )  RETURN boolean;
    FUNCTION "<=" ( l : t_state;  r : t_logic )  RETURN boolean;


    -- ***************************************************************************
    --
    --             PUBLIC OVERLOADED LOGICAL OPERATORS DECLARATIONS.
    --
    -- ***************************************************************************
    
    FUNCTION "and" ( l : t_logic;  r : t_logic )  RETURN t_state;
    FUNCTION "and" ( l : t_state;  r : t_logic )  RETURN t_state;
    FUNCTION "and" ( l : t_logic;  r : t_state )  RETURN t_state;
    FUNCTION "and" ( l : t_state;  r : t_state )  RETURN t_state;

    FUNCTION "or" ( l : t_logic;  r : t_logic )  RETURN t_state;
    FUNCTION "or" ( l : t_state;  r : t_logic )  RETURN t_state;
    FUNCTION "or" ( l : t_logic;  r : t_state )  RETURN t_state;
    FUNCTION "or" ( l : t_state;  r : t_state )  RETURN t_state;

    FUNCTION "nand" ( l : t_logic;  r : t_logic )  RETURN t_state;
    FUNCTION "nand" ( l : t_state;  r : t_logic )  RETURN t_state;
    FUNCTION "nand" ( l : t_logic;  r : t_state )  RETURN t_state;
    FUNCTION "nand" ( l : t_state;  r : t_state )  RETURN t_state;

    FUNCTION "nor" ( l : t_logic;  r : t_logic )  RETURN t_state;
    FUNCTION "nor" ( l : t_state;  r : t_logic )  RETURN t_state;
    FUNCTION "nor" ( l : t_logic;  r : t_state )  RETURN t_state;
    FUNCTION "nor" ( l : t_state;  r : t_state )  RETURN t_state;

    FUNCTION "xor" ( l : t_logic;  r : t_logic )  RETURN t_state;
    FUNCTION "xor" ( l : t_state;  r : t_logic )  RETURN t_state;
    FUNCTION "xor" ( l : t_logic;  r : t_state )  RETURN t_state;
    FUNCTION "xor" ( l : t_state;  r : t_state )  RETURN t_state;


    -- ***************************************************************************
    --
    --                      PUBLIC TABLE (CONSTANT) DECLARATIONS.
    --
    -- ***************************************************************************
    
    ------------------------------------------------------------------------------
    --
    --   table name: f_state
    --   parameters:  
    --       IN      logic      --  t_logic        -- logic value
    --   
    --   RETURNs:    t_state    -- state value from logic value 
    --
    --   purpose:    Take a logic value, and RETURN the state
    --
    --   example:
    --               IF  (f_state( input_SIGNAL ) = '1') THEN
    --                   -- Some statements
    --               ELSIF (f_state( input_SIGNAL ) = '0') THEN
    --                   -- Some statements
    --               ELSE
    --                   -- Some statements
    --               END IF;
    --                  
    ------------------------------------------------------------------------------
    

    -- Given value RETURN the associated state
    CONSTANT f_state : f_state_t := (
             'X', -- U
             'X', -- D
             '0', -- Z0
             '1', -- Z1
             'X', -- ZDX
             'X', -- DZX
             'X', -- ZX
             '0', -- W0
             '1', -- W1
             '0', -- WZ0
             '1', -- WZ1
             'X', -- WDX
             'X', -- DWX
             'X', -- WZX
             'X', -- ZWX
             'X', -- WX
             '0', -- R0
             '1', -- R1
             '0', -- RW0
             '1', -- RW1
             '0', -- RZ0
             '1', -- RZ1
             'X', -- RDX
             'X', -- DRX
             'X', -- RZX
             'X', -- ZRX
             'X', -- RWX
             'X', -- WRX
             'X', -- RX
             '0', -- F0
             '1', -- F1
             '0', -- FR0
             '1', -- FR1
             '0', -- FW0
             '1', -- FW1
             '0', -- FZ0
             '1', -- FZ1
             'X', -- FDX
             'X', -- DFX
             'X', -- FZX
             'X', -- ZFX
             'X', -- FWX
             'X', -- WFX
             'X', -- FRX
             'X', -- RFX
             'X'  -- FX
    );

    ------------------------------------------------------------------------------
    --
    --   table name: f_strength
    --   parameters:  
    --       IN      logic      --  t_logic        -- logic value
    --   
    --   RETURNs:    t_strength -- strength value from logic value
    --
    --   purpose:    Take a logic value, and RETURN the strength portion of that value.
    --
    --   notes:      This FUNCTION will RETURN the strongest strength always.  For
    --               instance when given QR00 it will RETURN R as the strength.
    --
    --   example:
    --               IF  (f_strength( input_SIGNAL ) = 'U') THEN
    --                   -- Some statements
    --               ELSIF (f_strength( input_SIGNAL ) = 'F') THEN
    --                   -- Some statements
    --               ELSE
    --                   -- Some statements
    --               END IF;
    --                  
    ------------------------------------------------------------------------------

    -- Given logic value RETURN the strength
    CONSTANT f_strength : f_strength_t := (
             'F', -- U
             'U', -- D
             'Z', -- Z0
             'Z', -- Z1
             'Z', -- ZDX
             'Z', -- DZX
             'Z', -- ZX
             'W', -- W0
             'W', -- W1
             'W', -- WZ0
             'W', -- WZ1
             'W', -- WDX
             'W', -- DWX
             'W', -- WZX
             'W', -- ZWX
             'W', -- WX
             'R', -- R0
             'R', -- R1
             'R', -- RW0
             'R', -- RW1
             'R', -- RZ0
             'R', -- RZ1
             'R', -- RDX
             'R', -- DRX
             'R', -- RZX
             'R', -- ZRX
             'R', -- RWX
             'R', -- WRX
             'R', -- RX
             'F', -- F0
             'F', -- F1
             'F', -- FR0
             'F', -- FR1
             'F', -- FW0
             'F', -- FW1
             'F', -- FZ0
             'F', -- FZ1
             'F', -- FDX
             'F', -- DFX
             'F', -- FZX
             'F', -- ZFX
             'F', -- FWX
             'F', -- WFX
             'F', -- FRX
             'F', -- RFX
             'F'  -- FX
    );

    -- Given logic value RETURN the strength
    CONSTANT f_strengthL : f_strength_t := (
             'F', -- U
             'U', -- D
             'Z', -- Z0
             'Z', -- Z1
             'Z', -- ZDX
             'U', -- DZX
             'Z', -- ZX
             'W', -- W0
             'W', -- W1
             'Z', -- WZ0
             'Z', -- WZ1
             'W', -- WDX
             'U', -- DWX
             'W', -- WZX
             'Z', -- ZWX
             'W', -- WX
             'R', -- R0
             'R', -- R1
             'W', -- RW0
             'W', -- RW1
             'Z', -- RZ0
             'Z', -- RZ1
             'R', -- RDX
             'U', -- DRX
             'R', -- RZX
             'Z', -- ZRX
             'R', -- RWX
             'W', -- WRX
             'R', -- RX
             'F', -- F0
             'F', -- F1
             'R', -- FR0
             'R', -- FR1
             'W', -- FW0
             'W', -- FW1
             'Z', -- FZ0
             'Z', -- FZ1
             'F', -- FDX
             'U', -- DFX
             'F', -- FZX
             'Z', -- ZFX
             'F', -- FWX
             'W', -- WFX
             'F', -- FRX
             'R', -- RFX
             'F'  -- FX
    );

    -- Given logic value RETURN the strength
    CONSTANT f_strengthH : f_strength_t := (
             'F', -- U
             'U', -- D
             'Z', -- Z0
             'Z', -- Z1
             'U', -- ZDX
             'Z', -- DZX
             'Z', -- ZX
             'W', -- W0
             'W', -- W1
             'W', -- WZ0
             'W', -- WZ1
             'U', -- WDX
             'W', -- DWX
             'Z', -- WZX
             'W', -- ZWX
             'W', -- WX
             'R', -- R0
             'R', -- R1
             'R', -- RW0
             'R', -- RW1
             'R', -- RZ0
             'R', -- RZ1
             'U', -- RDX
             'R', -- DRX
             'Z', -- RZX
             'R', -- ZRX
             'W', -- RWX
             'R', -- WRX
             'R', -- RX
             'F', -- F0
             'F', -- F1
             'F', -- FR0
             'F', -- FR1
             'F', -- FW0
             'F', -- FW1
             'F', -- FZ0
             'F', -- FZ1
             'U', -- FDX
             'F', -- DFX
             'Z', -- FZX
             'F', -- ZFX
             'W', -- FWX
             'F', -- WFX
             'R', -- FRX
             'F', -- RFX
             'F'  -- FX
    );

    ------------------------------------------------------------------------------
    --
    --   table name: f_logic
    --   parameters:  
    --       IN      state        --  t_state        -- state value
    --       IN      strength     --  t_strength     -- strength value
    -- 
    --   RETURNs:    t_logic     -- logic value created from inputs 
    --
    --   purpose:    Takes a state and strength and RETURNs the logic value 
    --               associated with the pair.        
    --   example:
    --               output_SIGNAL <= f_logic( '1' )( 'F' ) AFTER delay;  -- F1
    --                  
    ------------------------------------------------------------------------------


    -- Given state/strength, RETURN logic value
    CONSTANT f_logic : f_log_con_t := (
            ( D ,     -- '0', 'U'
              Z0,     -- '0', 'Z'
              W0,     -- '0', 'W'
              R0,     -- '0', 'R'
              F0 ),   -- '0', 'F'
            ( D ,     -- '1', 'U'
              Z1,     -- '1', 'Z'
              W1,     -- '1', 'W'
              R1,     -- '1', 'R'
              F1 ),   -- '1', 'F'
            ( D ,     -- 'X', 'U'
              ZX,     -- 'X', 'Z'
              WX,     -- 'X', 'W'
              RX,     -- 'X', 'R'
              FX )    -- 'X', 'F'
        );

    -- Given strength/strength, RETURN logic value
    CONSTANT f_logic0 : f_logs_con_t := (
            ( D,      -- 'U', 'U'
              ZDX,    -- 'U', 'Z'
              WDX,    -- 'U', 'W'
              RDX,    -- 'U', 'R'
              FDX ),  -- 'U', 'F'
            ( ZDX,    -- 'Z', 'U'
              Z0,     -- 'Z', 'Z'
              WZ0,    -- 'Z', 'W'
              RZ0,    -- 'Z', 'R'
              FZ0 ),  -- 'Z', 'F'
            ( WDX,    -- 'W', 'U'
              WZ0,    -- 'W', 'Z'
              W0,     -- 'W', 'W'
              RW0,    -- 'W', 'R'
              FW0 ),  -- 'W', 'F'
            ( RDX,    -- 'R', 'U'
              RZ0,    -- 'R', 'Z'
              RW0,    -- 'R', 'W'
              R0,     -- 'R', 'R'
              FR0 ),  -- 'R', 'F'
            ( FDX,    -- 'F', 'U'
              FZ0,    -- 'F', 'Z'
              FW0,    -- 'F', 'W'
              FR0,    -- 'F', 'R'
              F0 )    -- 'F', 'F'
        );

    -- Given strength/strength, RETURN logic value
    CONSTANT f_logic1 : f_logs_con_t := (
            ( D,      -- 'U', 'U'
              DZX,    -- 'U', 'Z'
              DWX,    -- 'U', 'W'
              DRX,    -- 'U', 'R'
              DFX ),  -- 'U', 'F'
            ( DZX,    -- 'Z', 'U'
              Z1,     -- 'Z', 'Z'
              WZ1,    -- 'Z', 'W'
              RZ1,    -- 'Z', 'R'
              FZ1 ),  -- 'Z', 'F'
            ( DWX,    -- 'W', 'U'
              WZ1,    -- 'W', 'Z'
              W1,     -- 'W', 'W'
              RW1,    -- 'W', 'R'
              FW1 ),  -- 'W', 'F'
            ( DRX,    -- 'R', 'U'
              RZ1,    -- 'R', 'Z'
              RW1,    -- 'R', 'W'
              R1,     -- 'R', 'R'
              FR1 ),  -- 'R', 'F'
            ( DFX,    -- 'F', 'U'
              FZ1,    -- 'F', 'Z'
              FW1,    -- 'F', 'W'
              FR1,    -- 'F', 'R'
              F1 )    -- 'F', 'F'
        );

    -- Given strength/strength, RETURN logic value
    CONSTANT f_logicX : f_logs_con_t := (
            ( D,      -- 'U', 'U'
              DZX,    -- 'U', 'Z'
              DWX,    -- 'U', 'W'
              DRX,    -- 'U', 'R'
              DFX ),  -- 'U', 'F'
            ( ZDX,    -- 'Z', 'U'
              ZX,     -- 'Z', 'Z'
              ZWX,    -- 'Z', 'W'
              ZRX,    -- 'Z', 'R'
              ZFX ),  -- 'Z', 'F'
            ( WDX,    -- 'W', 'U'
              WZX,    -- 'W', 'Z'
              WX,     -- 'W', 'W'
              WRX,    -- 'W', 'R'
              WFX ),  -- 'W', 'F'
            ( RDX,    -- 'R', 'U'
              RZX,    -- 'R', 'Z'
              RWX,    -- 'R', 'W'
              RX,     -- 'R', 'R'
              RFX ),  -- 'R', 'F'
            ( FDX,    -- 'F', 'U'
              FZX,    -- 'F', 'Z'
              FWX,    -- 'F', 'W'
              FRX,    -- 'F', 'R'
              FX )    -- 'F', 'F'
        );

    -- ***************************************************************************
    --
    --                      TECHNOLOGY LOOKUP TABLES
    --
    -- ***************************************************************************

    ------------------------------------------------------------------------------
    --
    --   table name: f_ecl, f_cmos, f_nmos, f_ttl, f_ttloc
    --   parameters:  
    --       IN      state     --  t_state          -- state value
    --  
    --   RETURNs:    t_logic  -- logic value created from input and technology.
    --
    --   purpose:    Take a state value and the implicit techology of a SIGNAL and RETURN
    --               the logic value associated with an output of that technology.
    --                  
    --   example:
    --               new_state     := f_and( inp1 )( inp2 );
    --               output_SIGNAL <= f_ttl( new_state ) AFTER delay;
    --                  
    ------------------------------------------------------------------------------

    -- Technology lookup tables
    CONSTANT f_ecl : f_specific_t :=  
         ( R0,     -- '0' 
           F1,     -- '1' 
           RFX     -- 'X' 
         ); 

    CONSTANT f_cmos : f_specific_t :=  
         ( F0,     -- '0' 
           F1,     -- '1' 
           FX      -- 'X' 
         ); 

    CONSTANT f_nmos : f_specific_t :=  
         ( F0,     -- '0' 
           R1,     -- '1' 
           FRX     -- 'X' 
         ); 

    CONSTANT f_ttl : f_specific_t :=  
         ( F0,     -- '0' 
           F1,     -- '1' 
           FX      -- 'X' 
         ); 

    CONSTANT f_ttloc : f_specific_t :=  
         ( F0,     -- '0' 
           ZX,     -- '1' 
           FZX     -- 'X' 
         ); 

    ------------------------------------------------------------------------------
    --
    --   table name: f_tech
    --   parameters:  
    --       IN      state     --  t_state          -- state value
    --       IN      tech      --  t_technology     -- technology value
    --  
    --   RETURNs:    t_logic  -- logic value created from inputs 
    --
    --   purpose:    Takes a state value and the techology of a SIGNAL and RETURNs
    --               the logic value associated with an output of that technology.
    --                  
    --   example:
    --               new_state     := f_and( inp1 )( inp2 );
    --               output_SIGNAL <= f_tech( new_state )( ttl ) AFTER delay;
    --                  
    ------------------------------------------------------------------------------

    -- Given state and technology, RETURN appropriate output value
    CONSTANT f_tech : f_tech_t := ( 
            ( R0,     -- '0', ecl
              F0,     -- '0', cmos 
              F0,     -- '0', nmos 
              F0,     -- '0', ttl  
              F0 ),   -- '0', ttloc

            ( F1,     -- '1', ecl
              F1,     -- '1', cmos
              R1,     -- '1', nmos
              F1,     -- '1', ttl 
              ZX ),   -- '1', ttloc

            ( RFX,    -- 'X', ecl
              FX,     -- 'X', cmos
              FRX,    -- 'X', nmos
              FX,     -- 'X', ttl 
              FZX )   -- 'X', ttloc

        );

    ------------------------------------------------------------------------------
    --
    --   table name: f_convz
    --   parameters:  
    --       IN      logic      --  t_logic        -- logic value
    --   
    --   RETURNs:    t_logic   -- logic value created from the input.
    --
    --   purpose:    Take a logic value, and convert it to the appropriate value with
    --               high impedance strength.  (Z)
    --
    --   notes:      This table is useful IN calculating the new state of a tristate
    --               node that has just had the control SIGNAL turned off.
    --
    --   example:
    --               output_SIGNAL <= f_convz( input_SIGNAL ) AFTER delay;
    --                  
    ------------------------------------------------------------------------------

    -- Given logic value, RETURN appropriate high-impedance tristate logic value
    CONSTANT f_convz : f_logic_t := (
              ZX, -- U
              D, -- D
              Z0, -- Z0
              Z1, -- Z1
              ZDX, -- ZDX
              DZX, -- DZX
              ZX, -- ZX
              Z0, -- W0
              Z1, -- W1
              Z0, -- WZ0
              Z1, -- WZ1
              ZDX, -- WDX
              DZX, -- DWX
              ZX, -- WZX
              ZX, -- ZWX
              ZX, -- WX
              Z0, -- R0
              Z1, -- R1
              Z0, -- RW0
              Z1, -- RW1
              Z0, -- RZ0
              Z1, -- RZ1
              ZDX, -- RDX
              DZX, -- DRX
              ZX, -- RZX
              ZX, -- ZRX
              ZX, -- RWX
              ZX, -- WRX
              ZX, -- RX
              Z0, -- F0
              Z1, -- F1
              Z0, -- FR0
              Z1, -- FR1
              Z0, -- FW0
              Z1, -- FW1
              Z0, -- FZ0
              Z1, -- FZ1
              ZDX, -- FDX
              DZX, -- DFX
              ZX, -- FZX
              ZX, -- ZFX
              ZX, -- FWX
              ZX, -- WFX
              ZX, -- FRX
              ZX, -- RFX
              ZX  -- FX
    );

    ------------------------------------------------------------------------------
    --
    --   FUNCTION name: f_convu
    --   parameters:  
    --       IN         newval    --  t_logic       -- new state value  (ONLY STATE
    --                                                  information is used)
    --       IN         lastval   --  t_logic       -- old state value
    --       IN         tech      --  t_technology   -- technology of gate
    -- 
    --   RETURNs:       t_logic  -- strongest possible state 
    --
    --   purpose:       This FUNCTION calculates the state that should be output
    --                  from a tristateable output when the gating SIGNAL is of an 
    --                  unknown value. This FUNCTION basically calculates the
    --                  strongest possible state that this gate could put out. 
    --                 
    ------------------------------------------------------------------------------
    
    FUNCTION f_convu ( newval    : IN t_logic;
                       lastval   : IN t_logic; 
                       tech      : IN t_technology ) RETURN t_logic;

    
    ------------------------------------------------------------------------------
    --
    --   FUNCTION name: f_convx
    --   parameters:  
    --       IN         newval    --  t_logic       -- new state value  (ONLY STATE
    --                                                  information is used)
    --       IN         lastval   --  t_logic       -- old state value
    --       IN         tech      --  t_technology   -- technology of gate
    -- 
    --   RETURNs:       t_logic  -- strongest possible state 
    --
    --   purpose:       This FUNCTION calculates the state that should be output
    --                  from a tristateable output when the gating SIGNAL is of an 
    --                  unknown value. This FUNCTION basically calculates the
    --                  strongest possible state that this gate could put out. 
    --                 
    ------------------------------------------------------------------------------
    
    FUNCTION f_convx ( newval    : IN t_logic;
                       lastval   : IN t_logic; 
                       tech      : IN t_technology ) RETURN t_logic;

    
    ------------------------------------------------------------------------------
    --
    --   FUNCTION name: f_uxfr
    --   parameters:  
    --       IN         g    --  t_logic
    --       IN         src  --  t_logic
    --       IN         drn  --  t_logic
    -- 
    --   RETURNs:       t_logic  -- to drive Drn output
    --
    --   purpose:       This FUNCTION calculates the output value of Drn based on the
    --                  current values of G, Src, and Drn.
    --                  IF g is on then
    --                      the output to Drn is Src
    --                  ELSIF g is off then
    --                      the output to Drn is f_convz(Drn)
    --                  ELSE
    --                      the output to Drn is the interval between Src and f_convz(Drn)
    --                  END IF;
    --                 
    ------------------------------------------------------------------------------
    
    FUNCTION f_uxfr ( g, src, drn  : IN t_logic ) RETURN t_logic;


    ------------------------------------------------------------------------------
    --
    --   FUNCTION name: f_resistor
    --   parameters:  
    --       IN         s    --  t_logic
    --       IN         size --  natural
    -- 
    --   RETURNs:       t_logic  -- to drive d output
    --
    --   purpose:       This FUNCTION calculates the output value of d based on the
    --                  current values of s. The strength of s is decreased by size level.
    --
    ------------------------------------------------------------------------------
    
    FUNCTION f_resistor ( s  : IN t_logic; size : IN natural ) RETURN t_logic;


    ------------------------------------------------------------------------------
    --
    --   FUNCTION name: f_ceil
    --   parameters:  
    --       IN         s    --  t_logic
    --       IN         size --  t_strength
    -- 
    --   RETURNs:       t_logic  -- to drive D output
    --
    --   purpose:       This FUNCTION calculates the output value of D based on the
    --                  current values of S. The strength of s is decreased by size level.
    --
    ------------------------------------------------------------------------------
    
    FUNCTION f_ceil ( s  : IN t_logic; size : IN t_strength ) RETURN t_logic;


    ------------------------------------------------------------------------------
    --
    --   FUNCTION name: f_cap
    --   parameters:  
    --       IN         s    --  t_logic
    --       IN         size --  t_strength
    -- 
    --   RETURNs:       t_logic  -- to drive d output
    --
    --   purpose:       This FUNCTION implements the capacitor.
    --                  IF the strength of s is larger or equal to the
    --                     size of the capacitor THEN
    --                      RETURN s;
    --                  ELSE
    --                      RETURN the value charged up IN the capacitor.
    --                  END IF;
    --
    --
    ------------------------------------------------------------------------------
    
    FUNCTION f_cap ( s  : IN t_logic; size : IN t_strength ) RETURN t_logic;


    -- ***************************************************************************
    --
    --                           LOGIC LOOKUP TABLES
    --
    -- ***************************************************************************


    ------------------------------------------------------------------------------
    --
    --   table name: f_and
    --   parameters:  
    --       IN      state1    --  t_state     -- state value
    --       IN      state2    --  t_state     -- state value
    --   
    --   RETURNs:    t_state   -- state value created from inputs 
    --
    --   purpose:    Computes (state1 and state2) and RETURNs its value.
    --                  
    --   notes:      The overloaded "and" operator can be obtained within the
    --               accompanying "std_ecl", "std_cmos", "std_nmos", "std_ttl"
    --               or "std_ttloc" packages.
    -- 
    --   example:
    --               ASSERT ( f_and( '1' )( '1' ) = '1' );
    --
    ------------------------------------------------------------------------------
    
    CONSTANT f_and : f_2_x_1 := ( 
            ( '0',   -- '0', '0'
              '0',   -- '0', '1', 
              '0' ), -- '0', 'X' 
            ( '0',   -- '1', '0'
              '1',   -- '1', '1'
              'X' ), -- '1', 'X'
            ( '0',   -- 'X', '0'
              'X',   -- 'X', '1'
              'X' )  -- 'X', 'X'
        );


    ------------------------------------------------------------------------------
    --
    --   table name: f_or 
    --   parameters:  
    --       IN      state1    --  t_state     -- state value
    --       IN      state2    --  t_state     -- state value
    --   
    --   RETURNs:    t_state   -- state value created from inputs 
    --
    --   purpose:    Computes (state1 or state2) and RETURNs its value.
    --                  
    --   notes:      The overloaded "or" operator can be obtained within the
    --               accompanying "std_ecl", "std_cmos", "std_nmos", "std_ttl"
    --               or "std_ttloc" packages.
    -- 
    --   example:
    --               ASSERT ( f_or( '1' )( '0' ) = '1' );
    --
    ------------------------------------------------------------------------------
    
    CONSTANT f_or : f_2_x_1 := ( 
            ( '0',   -- '0', '0'
              '1',   -- '0', '1', 
              'X' ), -- '0', 'X' 
            ( '1',   -- '1', '0'
              '1',   -- '1', '1'
              '1' ), -- '1', 'X'
            ( 'X',   -- 'X', '0'
              '1',   -- 'X', '1'
              'X' )  -- 'X', 'X'
        );
                                                    

    ------------------------------------------------------------------------------
    --
    --   table name: f_nand
    --   parameters:  
    --       IN      state1    --  t_state     -- state value
    --       IN      state2    --  t_state     -- state value
    --   
    --   RETURNs:    t_state   -- state value created from inputs 
    --
    --   purpose:    Computes (state1 nand state2) and RETURNs its value.
    --                  
    --   notes:      The overloaded "nand" operator can be obtained within the
    --               accompanying "std_ecl", "std_cmos", "std_nmos", "std_ttl"
    --               or "std_ttloc" packages.
    -- 
    --   example:
    --               ASSERT ( f_nand( '1' )( '1' ) = '0' );
    --
    ------------------------------------------------------------------------------
    
    CONSTANT f_nand : f_2_x_1 := ( 
            ( '1',   -- '0', '0'
              '1',   -- '0', '1', 
              '1' ), -- '0', 'X' 
            ( '1',   -- '1', '0'
              '0',   -- '1', '1'
              'X' ), -- '1', 'X'
            ( '1',   -- 'X', '0'
              'X',   -- 'X', '1'
              'X' )  -- 'X', 'X'
        );

    
    ------------------------------------------------------------------------------
    --
    --   table name: f_nor 
    --   parameters:  
    --       IN      state1    --  t_state     -- state value
    --       IN      state2    --  t_state     -- state value
    --   
    --   RETURNs:    t_state   -- state value created from inputs 
    --
    --   purpose:    Computes (state1 nor state2) and RETURNs its value.
    --                  
    --   notes:      The overloaded "nor" operator can be obtained within the
    --               accompanying "std_ecl", "std_cmos", "std_nmos", "std_ttl"
    --               or "std_ttloc" packages.
    -- 
    --   example:
    --               ASSERT ( f_nor( '1' )( '0' ) = '0' );
    --
    ------------------------------------------------------------------------------
    
    CONSTANT f_nor : f_2_x_1 := ( 
            ( '1',   -- '0', '0'
              '0',   -- '0', '1', 
              'X' ), -- '0', 'X' 
            ( '0',   -- '1', '0'
              '0',   -- '1', '1'
              '0' ), -- '1', 'X'
            ( 'X',   -- 'X', '0'
              '0',   -- 'X', '1'
              'X' )  -- 'X', 'X'
        );
    

    ------------------------------------------------------------------------------
    --
    --   table name: f_xor 
    --   parameters:  
    --       IN      state1    --  t_state     -- state value
    --       IN      state2    --  t_state     -- state value
    --   
    --   RETURNs:    t_state   -- state value created from inputs 
    --
    --   purpose:    Computes (state1 xor state2) and RETURNs its value.
    --                  
    --   notes:      The overloaded "xor" operator can be obtained within the
    --               accompanying "std_ecl", "std_cmos", "std_nmos", "std_ttl"
    --               "std_ttloc" packages.
    -- 
    --   example:
    --               ASSERT ( f_xor( '1' )( '1' ) = '0' );
    --
    ------------------------------------------------------------------------------
    
    CONSTANT f_xor : f_2_x_1 := ( 
            ( '0',   -- '0', '0'
              '1',   -- '0', '1', 
              'X' ), -- '0', 'X' 
            ( '1',   -- '1', '0'
              '0',   -- '1', '1'
              'X' ), -- '1', 'X'
            ( 'X',   -- 'X', '0'
              'X',   -- 'X', '1'
              'X' )  -- 'X', 'X'
        );


    ------------------------------------------------------------------------------
    --
    --   table name: f_not
    --   parameters:  
    --       IN      state     --  t_state     -- state value
    -- 
    --   RETURNs:    t_state   -- state value created from input 
    --
    --   purpose:    Computes (not state) and RETURNs its value.
    --                 
    --   notes:      The overloaded "not" operator can be obtained within the
    --               accompanying "std_ecl", "std_cmos", "std_nmos", "std_ttl"
    --               or "std_ttloc" packages.
    --
    --   example:
    --               ASSERT ( f_not( '1' ) = '0' );
    --
    ------------------------------------------------------------------------------
    
    CONSTANT f_not : f_1_x_1 := ( 
              '1',   --  '0'
              '0',   --  '1'
              'X'    --  'X' 
        );
    
    -- ***************************************************************************
    --
    --                          CLOCK EDGE LOOKUP TABLES
    --
    -- ***************************************************************************

    ------------------------------------------------------------------------------
    --
    --   table name: f_rising_edge
    --   parameters:  
    --       IN      logic      --  t_logic        -- logic value
    --   
    --   RETURNs:    boolean    -- TRUE IF state is '1' ('0'), FALSE otherwise.
    --
    --   purpose:    Determine whether the state of the SIGNAL value passed indicates
    --               whether the value is the result of a rising (falling) edge.
    --
    --   example:
    --               IF  (f_rising_edge( clk )) THEN
    --                   -- Check the 'setup' constraints on the inputs.
    --               END IF;
    --                  
    ------------------------------------------------------------------------------
    
    -- Given value, RETURN true for rising edge
    CONSTANT f_rising_edge : f_boolean_t := (
             false, -- U
             false, -- D
             false, -- Z0
             true,  -- Z1
             false, -- ZDX
             false, -- DZX
             false, -- ZX
             false, -- W0
             true,  -- W1
             false, -- WZ0
             true,  -- WZ1
             false, -- WDX
             false, -- DWX
             false, -- WZX
             false, -- ZWX
             false, -- WX
             false, -- R0
             true,  -- R1
             false, -- RW0
             true,  -- RW1
             false, -- RZ0
             true,  -- RZ1
             false, -- RDX
             false, -- DRX
             false, -- RZX
             false, -- ZRX
             false, -- RWX
             false, -- WRX
             false, -- RX
             false, -- F0
             true,  -- F1
             false, -- FR0
             true,  -- FR1
             false, -- FW0
             true,  -- FW1
             false, -- FZ0
             true,  -- FZ1
             false, -- FDX
             false, -- DFX
             false, -- FZX
             false, -- ZFX
             false, -- FWX
             false, -- WFX
             false, -- FRX
             false, -- RFX
             false  -- FX
    );

    ------------------------------------------------------------------------------
    --
    --   table name: f_falling_edge
    --   parameters:  
    --       IN      logic      --  t_logic        -- logic value
    --   
    --   RETURNs:    boolean    -- TRUE if state is '1' ('0'), FALSE otherwise.
    --
    --   purpose:    Determine whether the state of the SIGNAL value passed indicates
    --               whether the value is the result of a falling (falling) edge.
    --
    --   example:
    --               IF  (f_falling_edge( clk )) THEN
    --                   -- Check the 'setup' constraints on the inputs.
    --               END IF;
    --                  
    ------------------------------------------------------------------------------

    -- Given value, RETURN true for falling edge
    CONSTANT f_falling_edge : f_boolean_t := (
             false, -- U
             false, -- D
             true,  -- Z0
             false, -- Z1
             false, -- ZDX
             false, -- DZX
             false, -- ZX
             true,  -- W0
             false, -- W1
             true,  -- WZ0
             false, -- WZ1
             false, -- WDX
             false, -- DWX
             false, -- WZX
             false, -- ZWX
             false, -- WX
             true,  -- R0
             false, -- R1
             true,  -- RW0
             false, -- RW1
             true,  -- RZ0
             false, -- RZ1
             false, -- RDX
             false, -- DRX
             false, -- RZX
             false, -- ZRX
             false, -- RWX
             false, -- WRX
             false, -- RX
             true,  -- F0
             false, -- F1
             true,  -- FR0
             false, -- FR1
             true,  -- FW0
             false, -- FW1
             true,  -- FZ0
             false, -- FZ1
             false, -- FDX
             false, -- DFX
             false, -- FZX
             false, -- ZFX
             false, -- FWX
             false, -- WFX
             false, -- FRX
             false, -- RFX
             false  -- FX
    );


    -- ***************************************************************************
    --
    --                          TIMING UTILITY FUNCTIONS
    --
    -- ***************************************************************************
    
    ------------------------------------------------------------------------------
    --
    --   FUNCTION name: f_delay
    --   parameters:  
    --       IN         newlv   --  t_logic  -- new logic value
    --       IN         delay01 --  time      -- 0->1 delay value
    --       IN         delay10 --  time      -- 1->0 delay value
    -- 
    --   RETURNs:       The appropriate delay to be used, given the new value
    --                  and the 0-1 and 1-0 delays.
    --
    --   purpose:       Compute the appropriate delay to be used for a transition
    --                  on an output port.
    --
    ------------------------------------------------------------------------------
    
    FUNCTION f_delay(   newlv    : IN t_logic; 
                        delay01  : IN time;
                        delay10  : IN time) RETURN time;

    
    ------------------------------------------------------------------------------
    --
    --   FUNCTION name: f_zdelay
    --   parameters:  
    --       IN         oldlv   --  t_logic  -- old logic value
    --       IN         newlv   --  t_logic  -- new logic value
    --       IN         delay01 --  time      -- 0->1 delay value
    --       IN         delay10 --  time      -- 1->0 delay value
    --       IN         delayz0 --  time      -- z->0 delay value
    --       IN         delayz1 --  time      -- z->1 delay value
    --       IN         delay0z --  time      -- 0->z delay value
    --       IN         delay1z --  time      -- 1->z delay value
    --    
    --   RETURNs:       time    -- The appropriate delay to be used.
    --
    --   purpose:       Compute the delay to schedule for a tristate output based on 
    --                  the transistion of the output. If the output is going to a 1      
    --                  from 0 then use the 0->1 delay, which is delay01. If the output
    --                  is going from a 1->z value the use the delay1z value.  etc.
    --                 
    ------------------------------------------------------------------------------
    
    FUNCTION f_zdelay( oldlv     : IN t_logic;
                       newlv     : IN t_logic; 
                       delay01   : IN time; 
                       delay10   : IN time;
                       delayz0   : IN time;
                       delayz1   : IN time;
                       delay0z   : IN time;
                       delay1z   : IN time) RETURN time;

    
    ------------------------------------------------------------------------------
    --
    --   FUNCTION name: f_setup_check
    --   parameters:  
    --       SIGNAL IN  signal_to_check  --  t_logic  -- Signal to be checked for
    --                                                    setup violations.
    --       IN         Setup_Constraint --  time      -- 'Signal_to_Check' must have
    --                                                    been stable for at least
    --                                                    this amount of time.
    -- 
    --   RETURNs:       boolean         -- FALSE if setup time on input is too small.
    --
    --   purpose:       Checks to see that the setup time is not violated between the 
    --                  data and clock inputs.
    --
    --   example:                  
    --                 
    --                 -- 'Setup' checking on a positive-edge triggered clock.
    --                 setup_check: PROCESS( clk )
    --                 BEGIN
    --                 
    --                     -- Only perform the assertion after the rising edge.
    --                     IF  (f_rising_edge( clk )) THEN
    --                 
    --                         -- Verify that the setup constraint has not been violated.
    --                         ASSERT( f_setup_check( signal_to_check, setup_constraint ) )
    --                               REPORT "Setup check violation on 'signal_to_check'."
    --                               SEVERITY error;
    --                     END IF;
    --                 END PROCESS setup_check;
    --
    ------------------------------------------------------------------------------
    
    FUNCTION f_setup_check( SIGNAL signal_to_check  : IN t_logic; 
                                   setup_constraint : IN time ) RETURN boolean;

    
    ------------------------------------------------------------------------------
    --
    --   FUNCTION name: f_pulse_check
    --   parameters:  
    --       SIGNAL IN  signal_to_check --  t_logic  -- Signal to be checked for
    --                                                   a minimum pulse width.
    --       IN         Minimum_Width   --  time      -- Minimum width of the 
    --                                                   pulse for this SIGNAL.
    -- 
    --   RETURNs:       boolean        -- FALSE if pulse width on input is too small.
    --
    --   purpose:       Checks to make sure that a pulse on the 'signal_to_check' is not 
    --                  narrower than the time specified IN width.
    --
    --   example:
    --                 
    --                 -- Positive pulse checking.
    --                 pulse_check: PROCESS( clk )
    --                 BEGIN
    --                 
    --                     -- Only perform the assertion after the clock falls.
    --                     IF  (f_falling_edge( clk )) THEN
    --                 
    --                         -- Verify that the pulse width constraint has not been violated.
    --                         ASSERT( f_pulse_check( clk, minimum_width ) )
    --                               REPORT "Pulse check violation on 'clk'."
    --                               SEVERITY error;
    --                     END IF;
    --                 END PROCESS pulse_check;
    --
    ------------------------------------------------------------------------------
    
    FUNCTION f_pulse_check( SIGNAL signal_to_check : IN t_logic; 
                                   minimum_width   : IN time ) RETURN boolean;
    
    
    ------------------------------------------------------------------------------
    --
    --   FUNCTION name: f_hold_check
    --   parameters:  
    --       SIGNAL IN  signal_to_check --  t_logic  -- Signal to be checked for
    --                                                   Hold violations.
    --       IN         hold_time       --  time      -- Hold time constraint.
    --
    --   RETURNs:       boolean         -- FALSE if hold time on input is too small.
    --
    --   purpose:      Checks to see that the hold time of the input SIGNAL is long 
    --                 enough compared to the clock input.
    --            
    --                 
    --   notes:        To use this FUNCTION properly, you must wait for 'hold_time'
    --                 after the clock pulse and then call this FUNCTION.
    --
    --   example:
    --                 
    --                 -- Positive-edge triggered hold time checking PROCESS.
    --                 hold_check: PROCESS( clk'DELAYED( holdtime ) )
    --                 BEGIN
    --                 
    --                     -- Only perform the assertion after the rising edge.
    --                     IF  (f_rising_edge( clk'DELAYED( holdtime ) )) THEN
    --                 
    --                         -- Verify that the hold constraint has not been violated.
    --                         ASSERT( f_hold_check( signal_to_check, holdtime ) )
    --                               REPORT "Hold check violation on 'signal_to_check'."
    --                               SEVERITY error;
    --                     END IF;
    --                 END PROCESS hold_check;
    --
    ------------------------------------------------------------------------------
    
    FUNCTION f_hold_check( SIGNAL signal_to_check : IN t_logic; 
                                  hold_time       : IN time ) RETURN boolean;
    

    -- ***************************************************************************
    --
    --                      BUS RESOLUTION TABLE LOOKUP
    --
    -- ***************************************************************************
    --
    ------------------------------------------------------------------------------
    --
    --   table name: f_busres
    --   parameters:  
    --       IN      oldst   --  t_logic        -- old value
    --       IN      newst   --  t_logic        -- new value
    -- 
    --   RETURNs:    t_logic-- Combined value value created from inputs.
    --
    --   purpose:    Given two values RETURN the appropriate value 
    --   note:       In essense, these tables perform interval analysis RETURNing
    --               the largest interval spanning two given values
    ------------------------------------------------------------------------------

    -- Type definitions for table
    TYPE t_bus IS ARRAY(t_logic'LOW TO t_logic'HIGH) OF t_logic;
    TYPE t_bus_tab IS ARRAY(t_logic'LOW TO t_logic'HIGH) OF t_bus;

    CONSTANT f_busres : t_bus_tab := ( 
      ( U,  U,
        U,  U,  U,  U,  U,
        U,  U,  U,  U,  U,  U,  U,  U,  U,
        U,  U,  U,  U,  U,  U,  U,  U,  U,  U,  U,  U,  U,
        U,  U,  U,  U,  U,  U,  U,  U,  U,  U,  U,  U,  U,  U,  U,  U,  U ), -- U
      ( U,  D,
        Z0, Z1, ZDX, DZX, ZX,
        W0, W1, WZ0, WZ1, WDX, DWX, WZX, ZWX, WX,
        R0, R1, RW0, RW1, RZ0, RZ1, RDX, DRX, RZX, ZRX, RWX, WRX, RX,
        F0, F1, FR0, FR1, FW0, FW1, FZ0, FZ1, FDX, DFX, FZX, ZFX, FWX, WFX, FRX, RFX, FX ), -- D
      (  U,  Z0,
        Z0,  ZX,  Z0,  ZX,  ZX,
        W0,  W1, WZ0, ZWX, WZ0, ZWX, WZX, ZWX,  WX,
        R0,  R1, RW0, RW1, RZ0, ZRX, RZ0, ZRX, RZX, ZRX, RWX, WRX,  RX,
        F0,  F1, FR0, FR1, FW0, FW1, FZ0, ZFX, FZ0, ZFX, FZX, ZFX, FWX, WFX, FRX, RFX, FX ), -- Z0
      (  U,  Z1,
        ZX,  Z1,  ZX,  Z1,  ZX,
        W0,  W1, WZX, WZ1, WZX, WZ1, WZX, ZWX,  WX,
        R0,  R1, RW0, RW1, RZX, RZ1, RZX, RZ1, RZX, ZRX, RWX, WRX,  RX,
        F0,  F1, FR0, FR1, FW0, FW1, FZX, FZ1, FZX, FZ1, FZX, ZFX, FWX, WFX, FRX, RFX, FX ), -- Z1
      (  U, ZDX, 
        Z0,  ZX, ZDX,  ZX,  ZX,  
        W0,  W1, WZ0, ZWX, WDX, ZWX, WZX, ZWX,  WX,
        R0,  R1, RW0, RW1, RZ0, ZRX, RDX, ZRX, RZX, ZRX, RWX, WRX,  RX,
        F0,  F1, FR0, FR1, FW0, FW1, FZ0, ZFX, FDX, ZFX, FZX, ZFX, FWX, WFX, FRX, RFX, FX ), --ZDX
      (  U, DZX, 
        ZX,  Z1,  ZX, DZX,  ZX,
        W0,  W1, WZX, WZ1, WZX, DWX, WZX, ZWX,  WX,
        R0,  R1, RW0, RW1, RZX, RZ1, RZX, DRX, RZX, ZRX, RWX, WRX,  RX,
        F0,  F1, FR0, FR1, FW0, FW1, FZX, FZ1, FZX, DFX, FZX, ZFX, FWX, WFX, FRX, RFX, FX ), --DZX
      (  U,  ZX,  
        ZX,  ZX,  ZX,  ZX,  ZX,
        W0,  W1, WZX, ZWX, WZX, ZWX, WZX, ZWX,  WX,
        R0,  R1, RW0, RW1, RZX, ZRX, RZX, ZRX, RZX, ZRX, RWX, WRX,  RX,
        F0,  F1, FR0, FR1, FW0, FW1, FZX, ZFX, FZX, ZFX, FZX, ZFX, FWX, WFX, FRX, RFX, FX ), -- ZX
      (  U,  W0,
        W0,  W0,  W0,  W0,  W0,
        W0,  WX,  W0,  WX,  W0,  WX,  W0,  WX,  WX,
        R0,  R1, RW0, WRX, RW0, WRX, RW0, WRX, RW0, WRX, RWX, WRX,  RX,
        F0,  F1, FR0, FR1, FW0, WFX, FW0, WFX, FW0, WFX, FW0, WFX, FWX, WFX, FRX, RFX, FX ), -- W0
      (  U,  W1,
        W1,  W1,  W1,  W1,  W1,
        WX,  W1,  WX,  W1,  WX,  W1,  WX,  W1,  WX,
        R0,  R1, RWX, RW1, RWX, RW1, RWX, RW1, RWX, RW1, RWX, WRX,  RX,
        F0,  F1, FR0, FR1, FWX, FW1, FWX, FW1, FWX, FW1, FWX, FW1, FWX, WFX, FRX, RFX, FX ), -- W1
      (   U, WZ0,
        WZ0, WZX, WZ0, WZX, WZX,
         W0,  WX, WZ0,  WX, WZ0,  WX, WZX,  WX,  WX,
         R0,  R1, RW0, WRX, RZ0, WRX, RZ0, WRX, RZX, WRX, RWX, WRX,  RX,
         F0,  F1, FR0, FR1, FW0, WFX, FZ0, WFX, FZ0, WFX, FZX, WFX, FWX, WFX, FRX, RFX, FX ),--WZ0
      (   U, WZ1,
        ZWX, WZ1, ZWX, WZ1, ZWX,
         WX,  W1,  WX, WZ1,  WX, WZ1,  WX, ZWX,  WX,
         R0,  R1, RWX, RW1, RWX, RZ1, RWX, RZ1, RWX, ZRX, RWX, WRX,  RX,
         F0,  F1, FR0, FR1, FWX, FW1, FWX, FZ1, FWX, FZ1, FWX, ZFX, FWX, WFX, FRX, RFX, FX ),--WZ1
      (   U, WDX,
        WZ0, WZX, WDX, WZX, WZX,
         W0,  WX, WZ0,  WX, WDX,  WX, WZX,  WX,  WX,
         R0,  R1, RW0, WRX, RZ0, WRX, RDX, WRX, RZX, WRX, RWX, WRX,  RX,
         F0,  F1, FR0, FR1, FW0, WFX, FZ0, WFX, FDX, WFX, FZX, WFX, FWX, WFX, FRX, RFX, FX ),--WDX
      (   U, DWX,
        ZWX, WZ1, ZWX, DWX, ZWX,
         WX,  W1,  WX, WZ1,  WX, DWX,  WX, ZWX,  WX,
         R0,  R1, RWX, RW1, RWX, RZ1, RWX, DRX, RWX, ZRX, RWX, WRX,  RX,
         F0,  F1, FR0, FR1, FWX, FW1, FWX, FZ1, FWX, DFX, FWX, ZFX, FWX, WFX, FRX, RFX, FX ),--DWX
      (   U, WZX,
        WZX, WZX, WZX, WZX, WZX,
         W0,  WX, WZX,  WX, WZX,  WX, WZX,  WX,  WX,
         R0,  R1, RW0, WRX, RZX, WRX, RZX, WRX, RZX, WRX, RWX, WRX,  RX,
         F0,  F1, FR0, FR1, FW0, WFX, FZX, WFX, FZX, WFX, FZX, WFX, FWX, WFX, FRX, RFX, FX ),--WZX
      (   U, ZWX,
        ZWX, ZWX, ZWX, ZWX, ZWX,
         WX,  W1,  WX, ZWX,  WX, ZWX,  WX, ZWX,  WX,
         R0,  R1, RWX, RW1, RWX, ZRX, RWX, ZRX, RWX, ZRX, RWX, WRX,  RX,
         F0,  F1, FR0, FR1, FWX, FW1, FWX, ZFX, FWX, ZFX, FWX, ZFX, FWX, WFX, FRX, RFX, FX ),--ZWX
      (   U,  WX,
         WX,  WX,  WX,  WX,  WX,
         WX,  WX,  WX,  WX,  WX,  WX,  WX,  WX,  WX,
         R0,  R1, RWX, WRX, RWX, WRX, RWX, WRX, RWX, WRX, RWX, WRX,  RX,
         F0,  F1, FR0, FR1, FWX, WFX, FWX, WFX, FWX, WFX, FWX, WFX, FWX, WFX, FRX, RFX, FX ),--WX
      (   U,  R0,
         R0,  R0,  R0,  R0,  R0,  
         R0,  R0,  R0,  R0,  R0,  R0,  R0,  R0,  R0,
         R0,  RX,  R0,  RX,  R0,  RX,  R0,  RX,  R0,  RX,  R0,  RX,  RX,
         F0,  F1, FR0, RFX, FR0, RFX, FR0, RFX, FR0, RFX, FR0, RFX, FR0, RFX, FRX, RFX, FX ),--R0
      (   U,  R1,
         R1,  R1,  R1,  R1,  R1,
         R1,  R1,  R1,  R1,  R1,  R1,  R1,  R1,  R1,
         RX,  R1,  RX,  R1,  RX,  R1,  RX,  R1,  RX,  R1,  RX,  R1,  RX,
         F0,  F1, FRX, FR1, FRX, FR1, FRX, FR1, FRX, FR1, FRX, FR1, FRX, FR1, FRX, RFX, FX ),--R1
      (   U, RW0, 
        RW0, RW0, RW0, RW0, RW0,
        RW0, RWX, RW0, RWX, RW0, RWX, RW0, RWX, RWX,
         R0,  RX, RW0,  RX, RW0,  RX, RW0,  RX, RW0,  RX, RWX,  RX,  RX,
         F0,  F1, FR0, RFX, FW0, RFX, FW0, RFX, FW0, RFX, FW0, RFX, FWX, RFX, FRX, RFX, FX ),--RW0
      (   U, RW1,
        RW1, RW1, RW1, RW1, RW1,
        WRX, RW1, WRX, RW1, WRX, RW1, WRX, RW1, WRX,
         RX,  R1,  RX, RW1,  RX, RW1,  RX, RW1,  RX, RW1,  RX, WRX,  RX,
         F0,  F1, FRX, FR1, FRX, FW1, FRX, FW1, FRX, FW1, FRX, FW1, FRX, WFX, FRX, RFX, FX ),--RW1
      (   U, RZ0,
        RZ0, RZX, RZ0, RZX, RZX,
        RW0, RWX, RZ0, RWX, RZ0, RWX, RZX, RWX, RWX,
         R0,  RX, RW0,  RX, RZ0,  RX, RZ0,  RX, RZX,  RX, RWX,  RX,  RX,
         F0,  F1, FR0, RFX, FW0, RFX, FZ0, RFX, FZ0, RFX, FZX, RFX, FWX, RFX, FRX, RFX, FX ),--RZ0
      (   U, RZ1,
        ZRX, RZ1, ZRX, RZ1, ZRX,
        WRX, RW1, WRX, RZ1, WRX, RZ1, WRX, ZRX, WRX,
         RX,  R1,  RX, RW1,  RX, RZ1,  RX, RZ1,  RX, ZRX,  RX, WRX,  RX,
         F0,  F1, FRX, FR1, FRX, FW1, FRX, FZ1, FRX, FZ1, FRX, ZFX, FRX, WFX, FRX, RFX, FX ),--RZ1
      (   U, RDX,
        RZ0, RZX, RDX, RZX, RZX,
        RW0, RWX, RZ0, RWX, RDX, RWX, RZX, RWX, RWX,
         R0,  RX, RW0,  RX, RZ0,  RX, RDX,  RX, RZX,  RX, RWX,  RX,  RX,
         F0,  F1, FR0, RFX, FW0, RFX, FZ0, RFX, FDX, RFX, FZX, RFX, FWX, RFX, FRX, RFX, FX ),--RDX
      (   U, DRX,
        ZRX, RZ1, ZRX, DRX, ZRX,
        WRX, RW1, WRX, RZ1, WRX, DRX, WRX, ZRX, WRX,
         RX,  R1,  RX, RW1,  RX, RZ1,  RX, DRX,  RX, ZRX,  RX, WRX,  RX,
         F0,  F1, FRX, FR1, FRX, FW1, FRX, FZ1, FRX, DFX, FRX, ZFX, FRX, WFX, FRX, RFX, FX ),--DRX
      (   U, RZX,
        RZX, RZX, RZX, RZX, RZX,
        RW0, RWX, RZX, RWX, RZX, RWX, RZX, RWX, RWX,
         R0,  RX, RW0,  RX, RZX,  RX, RZX,  RX, RZX,  RX, RWX,  RX,  RX,
         F0,  F1, FR0, RFX, FW0, RFX, FZX, RFX, FZX, RFX, FZX, RFX, FWX, RFX, FRX, RFX, FX ),--RZX
      (   U, ZRX,
        ZRX, ZRX, ZRX, ZRX, ZRX,
        WRX, RW1, WRX, ZRX, WRX, ZRX, WRX, ZRX, WRX,
         RX,  R1,  RX, RW1,  RX, ZRX,  RX, ZRX,  RX, ZRX,  RX, WRX,  RX,
         F0,