USE work.tline_data_structure.all;
USE work.scientific.all;
PACKAGE tline_active_behaviors IS
--
--
-- function for linear transmission line driver
--
   FUNCTION drv_linear ( din :	bit;
			   res :  real ;
			   dout:  tline_node;
			   low :  real	    ;
			   hi  :  real	    ;
			   drv :  sig_mode   ) RETURN real;

--
--
-- transmit voltage function for transmission line segment
--
   FUNCTION transmit_volt( node_in :  tline_node    ;
			   int_in :   tline_traveler;
			   node_out:  tline_node    ;
			   res	   :  real	    ) RETURN real;


--
--
-- transmit far-end voltage function for transmission line segment
--
   FUNCTION transmit_vfe ( node_in :  tline_node ) RETURN real;


--
--
-- propagate function for transmission line segment
--
   FUNCTION propagate	 ( node_in :  tline_node    ;
			   int_in  :  tline_traveler
					  ) RETURN real;


--
--
-- branch current update function for transmission line segment
--
  FUNCTION update_current(node_in :  tline_node;
			  int_in  :   tline_traveler;
			  res	  :  real	) RETURN real;


--
--
-- function for transmission line driver using general voltage
--
   FUNCTION drive_general ( din : bit;
			   res :  real ;
			   dout:  tline_node;
			   low :  real	    ;
			   hi  :  real	    ;
			   drv :  sig_mode   ) RETURN real;


--
--
--
-- function for open drain transmission line FET driver
--
   FUNCTION drive_fet	 ( din :  bit	    ;
			   dout:  tline_node;
			   gm  :  real	    ;
			   wl  :  real	    ;
			   vt  :  real	    ;
			   vgs :  real	     ) RETURN real;


--
END tline_active_behaviors;


PACKAGE BODY tline_active_behaviors IS
--
--
-- function for linear transmission line driver
--
   FUNCTION drv_linear ( din :	bit;
			   res :  real ;
			   dout:  tline_node;
			   low :  real	    ;
			   hi  :  real	    ;
			   drv :  sig_mode   ) RETURN real IS

     VARIABLE gload: real      ; -- load conductance on driver output
     VARIABLE rload: real      ; -- load resistance  on driver output
     VARIABLE d_val: real      ; -- driver value

   BEGIN
	 gload	 := dout.g0 - (  1.0/(res + r_min) );
	 rload	 := 1.0/(gload + g_min);
	 IF (din = '0') THEN
	   d_val := low;
	 ELSE
	   d_val := hi;
	 END IF;
	 IF (drv = v) THEN
	   RETURN  d_val * ( rload/(rload + res) );
	 ELSE
	   RETURN  d_val * ( 1.0/dout.g0 );
	 END IF;
   END drv_linear;
--
--
-- transmit voltage function for transmission line segment
--
   FUNCTION transmit_volt( node_in :  tline_node    ;
			   int_in :   tline_traveler;
			   node_out:  tline_node    ;
			   res	   :  real	    ) RETURN real IS

     VARIABLE drive: real      ; -- returned
     VARIABLE gload: real      ; -- load conductance on driver output
     VARIABLE rload: real      ; -- load resistance  on driver output
     VARIABLE gamma: real      ; -- reflection coefficient

   BEGIN
       gload := node_out.g0 - ( 1.0/(res + r_min) );
       rload := 1.0/(gload + g_min);
       gamma := (rload - res)/(rload + res);
       RETURN  (node_in.volt - int_in.volt) * (1.0 + gamma);
   END transmit_volt;
--
--
-- transmit far-end voltage function for transmission line segment
--
   FUNCTION transmit_vfe ( node_in :  tline_node ) RETURN real IS

     VARIABLE drive: real; -- returned

   BEGIN
       drive := node_in.volt;
       RETURN drive;
   END transmit_vfe;
--
--
-- propagate function for transmission line segment
--
   FUNCTION propagate	 ( node_in :  tline_node    ;
			   int_in  :  tline_traveler
					  ) RETURN real IS

   BEGIN
       RETURN (node_in.volt - int_in.volt);

   END propagate;
--
--
-- branch current update function for transmission line segment
--
  FUNCTION update_current(node_in :  tline_node;
			  int_in  :   tline_traveler;
			  res	  :  real	) RETURN real IS

   BEGIN
       RETURN  ((node_in.volt - int_in.volt) - (int_in.volt))
				   /res;

   END update_current;
--
--
-- function for transmission line driver using general voltage
--
   FUNCTION drive_general ( din : bit;
			   res :  real ;
			   dout:  tline_node;
			   low :  real	    ;
			   hi  :  real	    ;
			   drv :  sig_mode   ) RETURN real IS

     VARIABLE gload: real      ; -- load conductance on driver output
     VARIABLE gdrv : real      ; -- load resistance  on driver output
     VARIABLE d_val: real      ; -- driver value
     VARIABLE v_int: real      ; -- voltage source intercept

   BEGIN
	 gdrv  := 1.0/(res + r_min);
	 gload := dout.g0 - g_max + g_min;
	 IF (din = '0') THEN
	   d_val := low;
	 ELSE
	   d_val := hi;
	 END IF;
	 IF (drv = v) THEN
	   v_int := d_val;
	 ELSE
	   v_int := d_val * res ;
	 END IF;
	 RETURN   ( dout.cur_fe + (dout.volt_fe * gload) +
		    (v_int * gdrv) )/(gload + gdrv);
   END drive_general;
--
--
--
-- function for open drain transmission line FET driver
--
   FUNCTION drive_fet	 ( din :  bit	    ;
			   dout:  tline_node;
			   gm  :  real	    ;
			   wl  :  real	    ;
			   vt  :  real	    ;
			   vgs :  real	     ) RETURN real IS

     VARIABLE gload: real; -- load conductance on driver output
     VARIABLE d_val: real; -- driver value
     VARIABLE kappa: real; -- voltage overdrive on fet
     VARIABLE beta : real; -- transconductance of fet
     VARIABLE bquad: real; -- b term in quadratic equation
     VARIABLE cquad: real; -- c term in quadratic equation
     VARIABLE disc : real; -- b**2 - c
     VARIABLE sat  : real; -- saturation current of fet

   BEGIN
	 gload := dout.g0 - g_max + g_min;
	 IF (din = '0') THEN
	  d_val := dout.volt_fe + (dout.cur_fe/gload);--hiz
	 ELSE
	  kappa :=  vgs - vt;
	  beta	:=  gm	* wl;
	  sat	:= (beta/2.0) * (kappa * kappa);
	  d_val := dout.volt_fe + (dout.cur_fe - sat)
				      /gload; -- saturated
	  IF (d_val < kappa) THEN
	    bquad := kappa + (gload/beta);
	    cquad := ( dout.cur_fe + (dout.volt_fe * gload) )
			   *(2.0/beta);
	    disc  := (bquad * bquad) - cquad;
	    d_val := bquad - sqrt(disc); -- linear
	  END IF;
	 END IF;
	 RETURN  d_val;
   END drive_fet;
--
END tline_active_behaviors;



